from functools import partial
from multiprocessing import Manager

from .progress_imap import progress_imap
from pandas.core.groupby.ops import _is_indexed_like
from pandas.util._decorators import doc

import dill

from .progress_imap import progress_udf_wrapper

DOC = 'Parallel analogue of the GroupBy.{func} method\nSee pandas GroupBy docstring for more ' \
      'information\nhttps://pandas.pydata.org/docs/reference/groupby.html'


def _do_group_apply(data, dill_func, workers_queue, args, kwargs):
    func = dill.loads(dill_func)
    result = progress_udf_wrapper(func, workers_queue, 1)(data[1], *args, **kwargs)
    return result, data[1].axes, 0


def _prepare_result(data):
    mutated = False
    result = list()
    for d in data:
        if not mutated and not _is_indexed_like(*d):
            mutated = True
        result.append(d[0])
    return result, mutated


def parallelize_groupby_apply(n_cpu=None, disable_pr_bar=False):
    @doc(DOC, func='apply')
    def p_apply(data, func, executor='processes', args=(), **kwargs):
        workers_queue = Manager().Queue()
        gr_count = data.ngroups
        iterator = iter(data)
        dill_func = dill.dumps(func)
        result = progress_imap(
            partial(_do_group_apply, dill_func=dill_func, workers_queue=workers_queue, args=args, kwargs=kwargs),
            iterator, workers_queue, total=gr_count, n_cpu=n_cpu, disable=disable_pr_bar, executor=executor,
            desc=func.__name__.upper()
        )
        result, mutated = _prepare_result(result)

        # due to a bug in the get_iterator method of the Basegrouper class
        # that was only fixed in pandas 1.4.0, earlier versions are not yet supported

        # pandas_version = get_pandas_version()
        #
        # if pandas_version < (1, 3):
        #     return data._wrap_applied_output(data.grouper._get_group_keys(), result,
        #                                      not_indexed_same=mutated or data.mutated)
        # elif pandas_version < (1, 4):
        #     return data._wrap_applied_output(data.grouper._get_group_keys(), data.grouper._get_group_keys(), result,
        #                                      not_indexed_same=mutated or data.mutated)
        return data._wrap_applied_output(data._selected_obj, result, not_indexed_same=mutated or data.mutated)

    return p_apply



from __future__ import annotations

from functools import partial
from multiprocessing import Manager

import pandas as pd
from pandas.util._decorators import doc

import dill

from .progress_imap import progress_imap
from .progress_imap import progress_udf_wrapper
from .tools import (
    get_split_data,
    get_split_size,
)

DOC = 'Parallel analogue of the pd.Series.{func} method\nSee pandas Series docstring for more ' \
      'information\nhttps://pandas.pydata.org/pandas-docs/stable/reference/series.html'


def _do_apply(data, dill_func, workers_queue, convert_dtype, args, kwargs):
    func = dill.loads(dill_func)
    if convert_dtype:
        return data.apply(progress_udf_wrapper(func, workers_queue, data.shape[0]), args=args, **kwargs)
    else:
        return data.astype(object).apply(progress_udf_wrapper(func, workers_queue, data.shape[0]), args=args, **kwargs)


def series_parallelize_apply(n_cpu=None, disable_pr_bar=False, show_vmem=False, split_factor=1):
    @doc(DOC, func='apply')
    def p_apply(data, func, executor='processes', convert_dtype=True, args=(), **kwargs):
        workers_queue = Manager().Queue()
        split_size = get_split_size(n_cpu, split_factor)
        tasks = get_split_data(data, 1, split_size)
        dill_func = dill.dumps(func)
        result = progress_imap(partial(_do_apply, convert_dtype=convert_dtype, dill_func=dill_func,
                                       workers_queue=workers_queue, args=args, kwargs=kwargs),
                               tasks, workers_queue, n_cpu=n_cpu, total=data.shape[0], disable=disable_pr_bar,
                               show_vmem=show_vmem, executor=executor, desc=func.__name__.upper())

        return pd.concat(result)

    return p_apply


def _do_map(data, dill_arg, workers_queue, na_action):
    func = dill.loads(dill_arg)
    def foo():
        return data.map(func, na_action=na_action)
    return progress_udf_wrapper(foo, workers_queue, 1)()


def series_parallelize_map(n_cpu=None, disable_pr_bar=False, show_vmem=False, split_factor=1):
    @doc(DOC, func='map')
    def p_map(data, arg, executor='threads', na_action=None):
        workers_queue = Manager().Queue()
        split_size = get_split_size(n_cpu, split_factor)
        tasks = get_split_data(data, 1, split_size)
        dill_arg = dill.dumps(arg)
        result = progress_imap(partial(_do_map, dill_arg=dill_arg,
                                       workers_queue=workers_queue, na_action=na_action),
                               tasks, workers_queue, n_cpu=n_cpu, total=split_size, disable=disable_pr_bar,
                               show_vmem=show_vmem, executor=executor, desc='map'.upper())

        return pd.concat(result)

    return p_map



from .parallel_series import series_parallelize_apply
from .parallel_series import series_parallelize_map
from .parallel_groupby import parallelize_groupby_apply
from .parallel_dataframe import parallelize_apply
from .parallel_dataframe import parallelize_replace
from .parallel_dataframe import parallelize_applymap
from .parallel_dataframe import parallelize_describe
from .parallel_dataframe import parallelize_nunique
from .parallel_dataframe import parallelize_mad
from .parallel_dataframe import parallelize_idxmax
from .parallel_dataframe import parallelize_idxmin
from .parallel_dataframe import parallelize_rank
from .parallel_dataframe import ParallelizeStatFunc
from .parallel_dataframe import ParallelizeStatFuncDdof
from .parallel_dataframe import ParallelizeMinCountStatFunc
from .parallel_dataframe import ParallelizeAccumFunc
from .parallel_dataframe import parallelize_quantile
from .parallel_dataframe import parallelize_mode
from .parallel_dataframe import parallelize_chunk_apply
from .parallel_dataframe import parallelize_merge
from .parallel_dataframe import parallelize_pct_change
from .parallel_dataframe import parallelize_isin
from .parallel_dataframe import parallelize_aggregate
from .parallel_dataframe import parallelize_corr
from .parallel_window import ParallelRolling
from .parallel_window import ParallelExpanding
from .parallel_window import ParallelEWM
from .parallel_window import ParallelRollingGroupby
from .parallel_window import ParallelExpandingGroupby
from .parallel_window import ParallelEWMGroupby
from .parallel_window import ParallelWindow




from __future__ import annotations

from functools import partial
from multiprocessing import cpu_count, Manager
import pandas as pd
from pandas.util._decorators import doc
import dill
from .progress_imap import progress_imap
from .progress_imap import progress_udf_wrapper
from .tools import get_split_data

DOC = 'Parallel analogue of the {func} method\nSee pandas DataFrame docstring for more ' \
      'information\nhttps://pandas.pydata.org/docs/reference/window.html'


class ParallelRolling:
    def __init__(self, n_cpu=None, disable_pr_bar=False, show_vmem=False, split_factor=1):
        self.n_cpu = n_cpu if n_cpu else cpu_count()
        self.disable_pr_bar = disable_pr_bar
        self.show_vmem = show_vmem
        self.split_factor = split_factor

    @staticmethod
    def _get_method(df, name, kwargs):
        return getattr(df.rolling(**kwargs), name)

    def do_method(self, df, workers_queue, name, serialized_flag, window_attr, args, kwargs):
        if name in ['apply', 'aggregate', 'agg']:
            if serialized_flag:
                # need to deserialize the function
                args = (dill.loads(args[0]),)
            elif isinstance(args[0], dict):
                _axis = df._get_axis_number(window_attr['axis'])
                if isinstance(df, pd.DataFrame):
                    func = {k: v for k, v in args[0].items() if k in df._get_axis(1 - _axis)}
                    args = (func,)

        def foo():
            method = self._get_method(df, name, window_attr)
            if name in ['aggregate', 'agg']:
                return method(*args, *kwargs['args'], **kwargs['kwargs'])
            return method(*args, **kwargs)

        return progress_udf_wrapper(foo, workers_queue, 1)()

    @staticmethod
    def _get_axis_and_offset(data):
        axis = 1
        offset = data.window
        if isinstance(data.obj, pd.DataFrame):
            axis = data.axis
            offset = 0
        return axis, offset

    def _get_split_data(self, data):
        axis, offset = self._get_axis_and_offset(data)
        return get_split_data(data.obj, axis, self.n_cpu * self.split_factor, offset=offset)

    def _get_total_tasks(self, data):
        axis = data.axis
        if isinstance(data.obj, pd.Series):
            axis = 1
        return min(self.n_cpu * self.split_factor, data.obj.shape[1 - axis])

    def _data_reduce(self, result, data):
        axis, offset = self._get_axis_and_offset(data)
        if offset:
            result = [result[0]] + [s[offset:] for s in result[1:]]
            return pd.concat(result, axis=1 - axis, ignore_index=True)
        return pd.concat(result, axis=1 - axis)

    @staticmethod
    def _func_serialize(func):
        if callable(func):
            return dill.dumps(func), True
        return func, False

    @staticmethod
    def _get_attributes(data):
        attributes = {attribute: getattr(data, attribute) for attribute in data._attributes}
        attributes.pop("_grouper", None)
        if attributes['win_type'] == 'freq':
            attributes['win_type'] = None
        return attributes

    def parallelize_method(self, data, name, executor, *args, **kwargs):
        attributes = self._get_attributes(data)
        workers_queue = Manager().Queue()
        serialized_flag = False
        if name in ['apply', 'aggregate', 'agg']:
            # if func is callable need to serialize it
            func, serialized_flag = self._func_serialize(args[0])
            if isinstance(func, dict):
                _axis = data.obj._get_axis_number(attributes['axis'])
                if _axis:
                    data.obj = data.obj.loc[[i for i in func.keys()]]
                else:
                    if len(func) == 1:
                        key = list(func.keys())
                        data.obj = data.obj[key[0]]
                    else:
                        data.obj = data.obj[[i for i in func.keys()]]
            args = (func,)
        tasks = self._get_split_data(data)
        total = self._get_total_tasks(data)
        result = progress_imap(
            partial(self.do_method, workers_queue=workers_queue, args=args, kwargs=kwargs, name=name,
                    window_attr=attributes, serialized_flag=serialized_flag),
            tasks, workers_queue, n_cpu=self.n_cpu, disable=self.disable_pr_bar, show_vmem=self.show_vmem,
            total=total, desc=name.upper(), executor=executor,
        )
        return self._data_reduce(result, data)

    def do_parallel(self, name):
        if name == 'mean':
            @doc(DOC, func=name)
            def p_mean(data, *args, executor='threads', engine=None, engine_kwargs=None, **kwargs):
                return self.parallelize_method(data, name, executor, *args, engine=engine, engine_kwargs=engine_kwargs,
                                               **kwargs)

            return p_mean

        if name == 'median':
            @doc(DOC, func=name)
            def p_median(data, executor='threads', engine=None, engine_kwargs=None, **kwargs):
                return self.parallelize_method(data, name, executor, engine=engine, engine_kwargs=engine_kwargs,
                                               **kwargs)

            return p_median

        if name == 'sum':
            @doc(DOC, func=name)
            def p_sum(data, *args, executor='threads', engine=None, engine_kwargs=None, **kwargs):
                return self.parallelize_method(data, name, executor, *args, engine=engine, engine_kwargs=engine_kwargs,
                                               **kwargs)

            return p_sum

        if name == 'min':
            @doc(DOC, func=name)
            def p_min(data, *args, executor='threads', engine=None, engine_kwargs=None, **kwargs):
                return self.parallelize_method(data, name, executor, *args, engine=engine, engine_kwargs=engine_kwargs,
                                               **kwargs)

            return p_min

        if name == 'max':
            @doc(DOC, func=name)
            def p_max(data, *args, executor='threads', engine=None, engine_kwargs=None, **kwargs):
                return self.parallelize_method(data, name, executor, *args, engine=engine, engine_kwargs=engine_kwargs,
                                               **kwargs)

            return p_max

        if name == 'std':
            @doc(DOC, func=name)
            def p_std(data, executor='threads', ddof=1, *args, engine=None, engine_kwargs=None, **kwargs):
                return self.parallelize_method(data, name, executor, ddof=ddof, *args, engine=engine,
                                               engine_kwargs=engine_kwargs,
                                               **kwargs)

            return p_std

        if name == 'var':
            @doc(DOC, func=name)
            def p_var(data, executor='threads', ddof=1, *args, engine=None, engine_kwargs=None, **kwargs):
                return self.parallelize_method(data, name, executor, ddof=ddof, *args, engine=engine,
                                               engine_kwargs=engine_kwargs,
                                               **kwargs)

            return p_var

        if name == 'sem':
            @doc(DOC, func=name)
            def p_sem(data, executor='threads', ddof=1, *args, **kwargs):
                return self.parallelize_method(data, name, executor, ddof=ddof, *args, **kwargs)

            return p_sem

        if name == 'skew':
            @doc(DOC, func=name)
            def p_skew(data, executor='threads', **kwargs):
                return self.parallelize_method(data, name, executor, **kwargs)

            return p_skew

        if name == 'kurt':
            @doc(DOC, func=name)
            def p_kurt(data, executor='threads', **kwargs):
                return self.parallelize_method(data, name, executor, **kwargs)

            return p_kurt

        if name == 'rank':
            @doc(DOC, func=name)
            def p_rank(data, executor='processes', method='average', ascending=True, pct=False, **kwargs):
                return self.parallelize_method(data, name, executor, method=method, ascending=ascending, pct=pct,
                                               **kwargs)

            return p_rank

        if name == 'quantile':
            @doc(DOC, func=name)
            def p_quantile(data, quantile, executor='threads', interpolation="linear", **kwargs):
                return self.parallelize_method(data, name, executor, quantile, interpolation=interpolation, **kwargs)

            return p_quantile

        if name == 'cov':
            @doc(DOC, func=name)
            def p_cov(data, executor='processes', other=None, pairwise=None, ddof=1, numeric_only=False):
                return self.parallelize_method(data, name, executor, other=other, pairwise=pairwise, ddof=ddof,
                                               numeric_only=numeric_only)

            return p_cov

        if name == 'apply':
            @doc(DOC, func=name)
            def p_apply(data, func, executor='processes', raw=False, engine=None, engine_kwargs=None, args=None,
                        kwargs=None):
                return self.parallelize_method(data, name, executor, func, raw=raw, engine=engine,
                                               engine_kwargs=engine_kwargs, args=args, kwargs=kwargs)

            return p_apply

        if name in ['aggregate', 'agg']:
            @doc(DOC, func=name)
            def p_aggregate(data, func, *args, executor='processes', **kwargs):
                return self.parallelize_method(data, name, executor, func, args=args, kwargs=kwargs)

            return p_aggregate


class ParallelWindow(ParallelRolling):
    @staticmethod
    def _get_attributes(data):
        attributes = {attribute: getattr(data, attribute) for attribute in data._attributes}
        attributes.pop("_grouper", None)
        return attributes

    def do_parallel(self, name):
        if name == 'mean':
            @doc(DOC, func=name)
            def p_mean(data, executor='threads', **kwargs):
                return self.parallelize_method(data, name, executor, **kwargs)

            return p_mean

        if name == 'sum':
            @doc(DOC, func=name)
            def p_sum(data, executor='threads', **kwargs):
                return self.parallelize_method(data, name, executor, **kwargs)

            return p_sum

        if name == 'std':
            @doc(DOC, func=name)
            def p_std(data, executor='threads', **kwargs):
                return self.parallelize_method(data, name, executor, **kwargs)

            return p_std

        if name == 'var':
            @doc(DOC, func=name)
            def p_var(data, executor='threads', **kwargs):
                return self.parallelize_method(data, name, executor, **kwargs)

            return p_var


class ParallelGroupbyMixin(ParallelRolling):

    def do_method(self, data, workers_queue, name, serialized_flag, window_attr, args, kwargs):
        if name in ['apply', 'aggregate', 'agg']:
            if serialized_flag:
                # need to deserialize the function
                args = (dill.loads(args[0]),)
            elif isinstance(args[0], dict):
                df = data[1]
                _axis = df._get_axis_number(window_attr['axis'])
                if isinstance(df, pd.DataFrame):
                    func = {k: v for k, v in args[0].items() if k in df._get_axis(1 - _axis)}
                    args = (func,)

        def foo():
            method = self._get_method(data[1], name, window_attr)
            if name in ['aggregate', 'agg']:
                result = method(*args, *kwargs['args'], **kwargs['kwargs'])
            else:
                result = method(*args, **kwargs)
            if isinstance(data[0], tuple):
                idx = [[i] for i in data[0]] + [result.index.tolist()]
            else:
                idx = [[data[0]], result.index.tolist()]
            result.index = pd.MultiIndex.from_product(idx)
            return result

        return progress_udf_wrapper(foo, workers_queue, 1)()

    def _get_split_data(self, data):
        return data._grouper.get_iterator(data.obj)

    def _get_total_tasks(self, data):
        return data._grouper.ngroups

    def _data_reduce(self, result, data):
        out = pd.concat(result)
        out.rename_axis(data._grouper.names + [data._grouper.axis.name], inplace=True)
        return out


class ParallelRollingGroupby(ParallelGroupbyMixin, ParallelRolling):
    pass


class ParallelExpanding(ParallelRolling):

    @staticmethod
    def _get_method(df, name, kwargs):
        return getattr(df.expanding(**kwargs), name)

    @staticmethod
    def _get_axis_and_offset(data):
        if isinstance(data.obj, pd.DataFrame):
            axis = data.axis
            offset = 0
        else:
            raise NotImplementedError('Parallel methods for Series objects are not implemented.')
        return axis, offset


class ParallelExpandingGroupby(ParallelGroupbyMixin, ParallelExpanding):
    pass


class ParallelEWM(ParallelRolling):

    @staticmethod
    def _get_axis_and_offset(data):
        if isinstance(data.obj, pd.DataFrame):
            axis = data.axis
            offset = 0
        else:
            raise NotImplementedError('Parallel methods for Series objects are not implemented.')
        return axis, offset

    @staticmethod
    def _get_method(df, name, kwargs):
        return getattr(df.ewm(**kwargs), name)


class ParallelEWMGroupby(ParallelGroupbyMixin, ParallelEWM):
    pass



import _thread as thread
import threading
from multiprocessing import cpu_count
from functools import wraps
from concurrent.futures import ThreadPoolExecutor

import platform
import signal
import os

import pandas as pd
import numpy as np
import time


def time_of_function(function):
    def wrapped(*args, **kwargs):
        start_time = time.time()
        res = function(*args, **kwargs)
        print('Time of function {} is {:.3f} s.'.format(function.__name__, time.time() - start_time))
        return res

    return wrapped


def get_pandas_version():
    major, minor = pd.__version__.split(".")[:2]
    return int(major), int(minor)


def _rank(mat):
    return np.argsort(np.argsort(mat, axis=0), axis=0)


def parallel_rank(mat, n_cpu):
    matrix_parts = np.array_split(mat, n_cpu, axis=1)
    with ThreadPoolExecutor(n_cpu) as pool:
        return np.hstack(list(pool.map(_rank, matrix_parts)))


def get_split_size(n_cpu, split_factor):
    if n_cpu is None:
        n_cpu = cpu_count()
    if split_factor is None:
        split_factor = 4
    return n_cpu * split_factor


def iterate_by_df(df, idx, axis, offset):
    if axis:
        for i in idx:
            start = max(0, i[0] - offset)
            yield df.iloc[start:i[-1] + 1]
    else:
        for i in idx:
            yield df.iloc[:, i[0]:i[-1] + 1]


def get_split_data(df, axis, split_size, offset=0):
    split_size = min(split_size, df.shape[1 - axis])
    idx_split = np.array_split(np.arange(df.shape[1 - axis]), split_size)
    tasks = iterate_by_df(df, idx_split, axis, offset)
    return tasks


def stop_function():
    if platform.system() == 'Windows':
        thread.interrupt_main()
    else:
        os.kill(os.getpid(), signal.SIGINT)


def stopit_after_timeout(s, raise_exception=True):
    def actual_decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            timer = threading.Timer(s, stop_function)
            try:
                timer.start()
                result = func(*args, **kwargs)
            except KeyboardInterrupt:
                msg = f'function \"{func.__name__}\" took longer than {s} s.'
                if raise_exception:
                    raise TimeoutError(msg)
                result = msg
            finally:
                timer.cancel()
            return result

        return wrapper

    return actual_decorator


def _wrapped_func(func, s, raise_exception, *args, **kwargs):
    return stopit_after_timeout(s, raise_exception=raise_exception)(func)(*args, **kwargs)



import time
from functools import partial
from itertools import count

import multiprocessing as mp
from threading import Thread

from tqdm.auto import tqdm

from .tools import _wrapped_func

from psutil import virtual_memory
from psutil._common import bytes2human


class ProgressBar(tqdm):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def close(self):
        super().close()
        if hasattr(self, 'disp'):
            if self.total and self.n < self.total:
                self.disp(bar_style='warning')


class MemoryProgressBar(tqdm):

    def refresh(self, **kwargs):
        super().refresh(**kwargs)
        if 70 <= self.n < 90:
            self.colour = 'orange'
        elif self.n >= 90:
            self.colour = 'red'
        else:
            self.colour = 'green'


class ProgressStatus:
    def __init__(self):
        self.next_update = 1
        self.last_update_t = time.perf_counter()
        self.last_update_val = 0


def progress_udf_wrapper(func, workers_queue, total):
    state = ProgressStatus()
    cnt = count(1)

    def wrapped_udf(*args, **kwargs):
        result = func(*args, **kwargs)
        updated = next(cnt)
        if updated == state.next_update:
            time_now = time.perf_counter()

            delta_t = time_now - state.last_update_t
            delta_i = updated - state.last_update_val

            state.next_update += max(int((delta_i / delta_t) * .25), 1)
            state.last_update_val = updated
            state.last_update_t = time_now
            workers_queue.put_nowait((1, delta_i))
        elif updated == total:
            workers_queue.put_nowait((1, updated - state.last_update_val))
        return result

    return wrapped_udf


def _process_status(bar_size, disable, show_vmem, desc, q):
    bar = ProgressBar(total=bar_size, disable=disable, desc=desc + ' DONE')
    vmem = virtual_memory()
    if show_vmem:
        vmem_pbar = MemoryProgressBar(range(100),
                                      bar_format="{desc}: {percentage:.1f}%|{bar}|  " + bytes2human(vmem.total),
                                      initial=vmem.percent, colour='green', position=1, desc='VMEM USAGE',
                                      disable=disable,
                                      )
        vmem_pbar.refresh()
    while True:
        flag, upd_value = q.get()
        if not flag:
            bar.close()
            if show_vmem:
                vmem_pbar.close()
            return
        bar.update(upd_value)
        if show_vmem:
            if time.time() - vmem_pbar.last_print_t >= 1:
                vmem = virtual_memory()
                vmem_pbar.update(vmem.percent - vmem_pbar.n)
                vmem_pbar.refresh()


def _do_parallel(func, tasks, initializer, initargs, n_cpu, total, disable, show_vmem,
                 q, executor, desc):
    if not n_cpu:
        n_cpu = mp.cpu_count()
    thread_ = Thread(target=_process_status, args=(total, disable, show_vmem, desc, q))
    thread_.start()
    if executor == 'threads':
        exc_pool = mp.pool.ThreadPool(n_cpu, initializer=initializer, initargs=initargs)
    else:
        exc_pool = mp.Pool(n_cpu, initializer=initializer, initargs=initargs)
    with exc_pool as p:
        result = list()
        iter_result = p.imap(func, tasks)
        while 1:
            try:
                result.append(next(iter_result))
            except StopIteration:
                break
    q.put((None, None))
    thread_.join()
    return result


def progress_imap(func, tasks, q, executor='threads', initializer=None, initargs=(), n_cpu=None, total=None,
                  disable=False, process_timeout=None, show_vmem=False, desc=''
                  ):
    if executor not in ['threads', 'processes']:
        raise ValueError('Invalid executor value specified. Must be one of the values: "threads", "processes"')
    try:
        if process_timeout:
            func = partial(_wrapped_func, func, process_timeout, True)
        result = _do_parallel(func, tasks, initializer, initargs, n_cpu, total, disable, show_vmem, q, executor, desc)
    except (KeyboardInterrupt, Exception):
        q.put((None, None))
        raise
    return result



from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from multiprocessing import cpu_count

from functools import partial
from multiprocessing import Manager
import numpy as np
import pandas as pd
from pandas._libs import lib
from pandas.core import nanops
from pandas.util._decorators import doc
from pandas._typing import (
    IndexLabel,
    Suffixes
)

from scipy.stats import spearmanr, kendalltau

import dill

from ._numba import _do_parallel_corr
from .progress_imap import progress_imap
from .progress_imap import progress_udf_wrapper
from .tools import (
    parallel_rank,
    get_split_data,
    get_split_size,
)

DOC = 'Parallel analogue of the DataFrame.{func} method\nSee pandas DataFrame docstring for more ' \
      'information\nhttps://pandas.pydata.org/docs/reference/frame.html '


def _do_apply(data, dill_func, workers_queue, axis, raw, result_type, args, kwargs):
    func = dill.loads(dill_func)
    return data.apply(progress_udf_wrapper(func, workers_queue, data.shape[1 - axis]), axis=axis, raw=raw,
                      result_type=result_type, args=args, **kwargs)


def parallelize_apply(n_cpu=None, disable_pr_bar=False, show_vmem=False, split_factor=1):
    @doc(DOC, func='apply')
    def p_apply(data, func, executor='processes', axis=0, raw=False, result_type=None, args=(), **kwargs):
        workers_queue = Manager().Queue()
        split_size = get_split_size(n_cpu, split_factor)
        tasks = get_split_data(data, axis, split_size)
        dill_func = dill.dumps(func)
        result = progress_imap(partial(_do_apply, axis=axis, raw=raw, result_type=result_type, dill_func=dill_func,
                                       workers_queue=workers_queue, args=args, kwargs=kwargs),
                               tasks, workers_queue, n_cpu=n_cpu, total=data.shape[1 - axis], disable=disable_pr_bar,
                               show_vmem=show_vmem, executor=executor, desc=func.__name__.upper())
        concat_axis = 0
        if result:
            if isinstance(result[0], pd.DataFrame):
                concat_axis = 1 - axis
        return pd.concat(result, axis=concat_axis)

    return p_apply


def _do_chunk_apply(data, dill_func, workers_queue, args, kwargs):
    func = dill.loads(dill_func)

    def foo():
        return func(data, *args, **kwargs)

    return progress_udf_wrapper(foo, workers_queue, 1)()


def parallelize_chunk_apply(n_cpu=None, disable_pr_bar=False, show_vmem=False, split_factor=1):
    def chunk_apply(data, func, executor='processes', axis=0, split_by_col=None,
                    concat_result=True, args=(), **kwargs):
        workers_queue = Manager().Queue()
        split_size = get_split_size(n_cpu, split_factor)
        if split_by_col:
            idx_split = np.array_split(data[split_by_col].unique(), split_size)
            group = data.groupby(split_by_col)
            tasks = (pd.concat([group.get_group(j) for j in i], copy=False) for i in idx_split)
        else:
            tasks = get_split_data(data, 1 - axis, split_size)
        dill_func = dill.dumps(func)
        result = progress_imap(partial(_do_chunk_apply, dill_func=dill_func,
                                       workers_queue=workers_queue, args=args, kwargs=kwargs),
                               tasks, workers_queue, n_cpu=n_cpu, total=split_size, disable=disable_pr_bar,
                               show_vmem=show_vmem, executor=executor, desc='chunk_apply'.upper())
        if concat_result:
            return pd.concat(result, axis=axis)
        else:
            return result

    return chunk_apply


def _np_pearson_corr(a, b=None):
    if b is None:
        return np.corrcoef(a, rowvar=False)
    return np.corrcoef(a, y=b)[0, 1]


def _run_in_pool(func, tasks, n_cpu, executor):
    chunk_size, extra = divmod(len(tasks), n_cpu)
    if extra:
        chunk_size += 1
    if executor == 'processes':
        pool = ProcessPoolExecutor(n_cpu)
    else:
        pool = ThreadPoolExecutor(n_cpu)
    with pool as p:
        result = list(p.map(func, tasks, chunksize=chunk_size))
    return result


def _pearson_corr(idx, mat, min_periods, isnan, isnan_flags, sums, sum_squareds):
    row, col = idx
    n = mat.shape[0]
    if isnan_flags[row] or isnan_flags[col]:
        valid = ~isnan[:, row] & ~isnan[:, col]
        if valid.sum() < min_periods:
            return np.nan
        else:
            return _np_pearson_corr(mat[valid, row], mat[valid, col])
    else:
        sum_xy = np.sum(mat[:, row] * mat[:, col])
        numerator = n * sum_xy - sums[row] * sums[col]
        denominator = np.sqrt(
            (n * sum_squareds[row] - sums[row] * sums[row]) * (n * sum_squareds[col] - sums[col] * sums[col]))
        return numerator / denominator if denominator != 0 else 0


def _parallel_pearson_corr(mat, min_periods, executor, n_cpu):
    comb = [(i, j) for i in range(mat.shape[1]) for j in range(i + 1, mat.shape[1])]
    isnan = np.isnan(mat)
    isnan_flags = isnan.sum(axis=0)
    sums = np.sum(mat, axis=0)
    sum_squareds = np.sum((mat * mat), axis=0)
    func = partial(_pearson_corr, mat=mat, min_periods=min_periods, isnan=isnan, isnan_flags=isnan_flags, sums=sums,
                   sum_squareds=sum_squareds)
    result = _run_in_pool(func, comb, n_cpu, executor)
    return result


def _spearman_corr(idx, mat, min_periods, isnan, isnan_flags, ranks):
    row, col = idx
    n = mat.shape[0]
    if isnan_flags[row] or isnan_flags[col]:
        valid = ~isnan[:, row] & ~isnan[:, col]
        if valid.sum() < min_periods:
            return np.nan
        else:
            return _scipy_spearman_corr(mat[valid, row], mat[valid, col])
    d = ranks[:, row] - ranks[:, col]
    sum_d_squared = np.sum(d * d)
    return 1 - (6 * sum_d_squared) / (n * (n ** 2 - 1))


def _parallel_spearman_corr(mat, min_periods, executor, n_cpu):
    comb = [(i, j) for i in range(mat.shape[1]) for j in range(i + 1, mat.shape[1])]
    isnan = np.isnan(mat)
    isnan_flags = isnan.sum(axis=0)
    ranks = parallel_rank(mat, n_cpu)
    func = partial(_spearman_corr, mat=mat, min_periods=min_periods, isnan=isnan, isnan_flags=isnan_flags, ranks=ranks)
    result = _run_in_pool(func, comb, n_cpu, executor)
    return result


def _kendall_tau(x, y):
    return kendalltau(x, y)[0]


def _scipy_spearman_corr(x, y):
    return spearmanr(x, y)[0]


def _do_corr(idx, mat, min_periods, corrf):
    a, b = mat[:, idx[0]], mat[:, idx[1]]
    valid = np.isfinite(a) & np.isfinite(b)
    if valid.sum() < min_periods:
        c = np.nan
    elif not valid.all():
        c = corrf(a[valid], b[valid])
    else:
        c = corrf(a, b)
    return c


def _parallel_do_corr(mat, min_periods, corrf, executor, n_cpu):
    comb = [(i, j) for i in range(mat.shape[1]) for j in range(i + 1, mat.shape[1])]
    func = partial(_do_corr, min_periods=min_periods, mat=mat,  corrf=corrf)
    result = _run_in_pool(func, comb, n_cpu, executor)
    return result


def parallelize_corr(n_cpu=None, disable_pr_bar=False, show_vmem=False, split_factor=1):
    if n_cpu is None:
        n_cpu = cpu_count()

    def p_corr(data, method='pearson', min_periods=1, numeric_only=False, executor='threads', engine=None):
        data = data._get_numeric_data() if numeric_only else data
        cols = data.columns
        idx = cols.copy()
        if method in ['pearson', 'spearman']:
            if not data.isna().any().any():
                if method == 'pearson':
                    result = _np_pearson_corr(data.values)
                elif method == 'spearman':
                    ranks = parallel_rank(data.values, n_cpu)
                    result = _np_pearson_corr(ranks)
                return pd.DataFrame(result, index=idx, columns=idx)
        mat = data.to_numpy(dtype=float, na_value=np.nan, copy=False)
        if engine == 'numba':
            return pd.DataFrame(_do_parallel_corr(mat, method=method, min_periods=min_periods), index=idx, columns=cols)
        if method == 'pearson':
            result = _parallel_pearson_corr(mat, min_periods, executor, n_cpu)
        elif method == 'spearman':
            result = _parallel_spearman_corr(mat, min_periods, executor, n_cpu)
        elif method == 'kendall':
            result = _parallel_do_corr(mat, min_periods, _kendall_tau, executor, n_cpu)
        elif callable(method):
            result = _parallel_do_corr(mat, min_periods, method, executor, n_cpu)
        else:
            raise ValueError(f'Unknown method {method}')

        corr_mat = np.zeros((data.shape[1], data.shape[1]))
        z = 0
        for j in range(0, data.shape[1]):
            for i in range(j + 1, data.shape[1]):
                corr_mat[i, j] = result[z]
                z += 1
        result_mat = corr_mat + corr_mat.transpose()
        np.fill_diagonal(result_mat, 1)
        return pd.DataFrame(result_mat, index=idx, columns=idx)

    return p_corr


def _do_aggregate(data, func, workers_queue, dilled_func, axis, args, kwargs):
    if dilled_func:
        func = dill.loads(func)
    if isinstance(func, dict):
        _axis = data._get_axis_number(axis)
        func = {k: v for k, v in func.items() if k in data._get_axis(1 - _axis)}

    def foo():
        return data.agg(func, axis=axis, *args, **kwargs)

    return progress_udf_wrapper(foo, workers_queue, 1)()


def parallelize_aggregate(n_cpu=None, disable_pr_bar=False, show_vmem=False, split_factor=1):
    @doc(DOC, func='aggregate')
    def p_aggregate(data, func, executor='threads', axis=0, args=(), **kwargs):
        workers_queue = Manager().Queue()
        split_size = get_split_size(n_cpu, split_factor)
        tasks = get_split_data(data, axis, split_size)
        dilled_func = False
        if callable(func):
            dilled_func = True
            func = dill.dumps(func)
        result = progress_imap(partial(_do_aggregate, axis=axis, func=func,
                                       workers_queue=workers_queue, dilled_func=dilled_func, args=args, kwargs=kwargs),
                               tasks, workers_queue, n_cpu=n_cpu, total=split_size, disable=disable_pr_bar,
                               show_vmem=show_vmem, executor=executor, desc='agg'.upper())
        concat_axis = 0
        if result:
            if isinstance(result[0], pd.DataFrame):
                concat_axis = 1 - axis
        return pd.concat(result, axis=concat_axis)

    return p_aggregate


def _do_replace(df, workers_queue, **kwargs):
    def foo():
        return df.replace(**kwargs)

    return progress_udf_wrapper(foo, workers_queue, 1)()


def parallelize_replace(n_cpu=None, disable_pr_bar=False, show_vmem=False, split_factor=1):
    @doc(DOC, func='replace')
    def p_replace(data, to_replace=None, value=lib.no_default, limit=None,
                  regex: bool = False, method: str | lib.NoDefault = lib.no_default):
        workers_queue = Manager().Queue()
        split_size = get_split_size(n_cpu, split_factor)
        tasks = get_split_data(data, 1, split_size)
        result = progress_imap(partial(_do_replace, to_replace=to_replace, value=value, limit=limit, regex=regex,
                                       method=method, workers_queue=workers_queue), tasks, workers_queue,
                               total=split_size, n_cpu=n_cpu, disable=disable_pr_bar, show_vmem=show_vmem,
                               desc='REPLACE')

        return pd.concat(result)

    return p_replace


def do_applymap(df, workers_queue, dill_func, na_action, kwargs):
    func = dill.loads(dill_func)

    return df.applymap(progress_udf_wrapper(func, workers_queue, df.size), na_action=na_action, **kwargs)


def parallelize_applymap(n_cpu=None, disable_pr_bar=False, show_vmem=False, split_factor=1):
    @doc(DOC, func='applymap')
    def p_applymap(data, func, na_action=None, **kwargs):
        workers_queue = Manager().Queue()
        split_size = get_split_size(n_cpu, split_factor)
        tasks = get_split_data(data, 1, split_size)
        dill_func = dill.dumps(func)
        result = progress_imap(
            partial(do_applymap, workers_queue=workers_queue, dill_func=dill_func, na_action=na_action,
                    kwargs=kwargs), tasks, workers_queue, n_cpu=n_cpu, disable=disable_pr_bar, show_vmem=show_vmem,
            total=data.size, executor='processes', desc='APPLYMAP')
        return pd.concat(result)

    return p_applymap


def do_describe(df, workers_queue, percentiles, include, exclude, datetime_is_numeric):
    if isinstance(df, pd.Series):
        df = df.to_frame()

    def foo():
        return df.describe(percentiles=percentiles, include=include, exclude=exclude,
                           datetime_is_numeric=datetime_is_numeric)

    return progress_udf_wrapper(foo, workers_queue, 1)()


def parallelize_describe(n_cpu=None, disable_pr_bar=False, show_vmem=False, split_factor=1):
    @doc(DOC, func='describe')
    def p_describe(data, percentiles=None, include=None, exclude=None,
                   datetime_is_numeric=False):
        workers_queue = Manager().Queue()
        split_size = get_split_size(n_cpu, split_factor)
        tasks = get_split_data(data, 0, split_size)
        result = progress_imap(
            partial(do_describe, workers_queue=workers_queue, percentiles=percentiles, include=include, exclude=exclude,
                    datetime_is_numeric=datetime_is_numeric), tasks, workers_queue, n_cpu=n_cpu, disable=disable_pr_bar,
            show_vmem=show_vmem, total=min(split_size, data.shape[1]), desc='DESCRIBE')
        return pd.concat(result, axis=1)

    return p_describe


def do_pct_change(df, workers_queue, periods, fill_method, limit, freq, kwargs):
    def foo():
        return df.pct_change(periods=periods, fill_method=fill_method, limit=limit, freq=freq, **kwargs)

    return progress_udf_wrapper(foo, workers_queue, 1)()


def parallelize_pct_change(n_cpu=None, disable_pr_bar=False, show_vmem=False, split_factor=1):
    @doc(DOC, func='pct_change')
    def p_pct_change(data, periods=1,
                     fill_method="pad",
                     limit=None,
                     freq=None,
                     **kwargs, ):
        axis = kwargs.get('axis', 0)
        workers_queue = Manager().Queue()
        split_size = get_split_size(n_cpu, split_factor)
        tasks = get_split_data(data, axis, split_size)
        result = progress_imap(
            partial(do_pct_change, workers_queue=workers_queue, periods=periods, fill_method=fill_method, limit=limit,
                    freq=freq, kwargs=kwargs), tasks, workers_queue, n_cpu=n_cpu, disable=disable_pr_bar,
            show_vmem=show_vmem, total=min(split_size, data.shape[1]), desc='PCT_CHANGE')
        return pd.concat(result, axis=1 - axis)

    return p_pct_change


def parallelize_nunique(n_cpu=None, disable_pr_bar=False, show_vmem=False, split_factor=1):
    @doc(DOC, func='nunique')
    def p_nunique(data, executor='threads', axis=0, dropna=True):
        return parallelize_apply(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar, show_vmem=show_vmem,
                                 split_factor=split_factor)(data, pd.Series.nunique, executor=executor,
                                                            axis=axis, dropna=dropna)

    return p_nunique


def do_mad(df, workers_queue, axis, skipna, level):
    def foo():
        return df.mad(axis=axis, skipna=skipna, level=level)

    return progress_udf_wrapper(foo, workers_queue, 1)()


def parallelize_mad(n_cpu=None, disable_pr_bar=False, show_vmem=False, split_factor=1):
    @doc(DOC, func='mad')
    def p_mad(data, axis=0, skipna=True, level=None):
        workers_queue = Manager().Queue()
        split_size = get_split_size(n_cpu, split_factor)
        tasks = get_split_data(data, axis, split_size)
        total = min(split_size, data.shape[1 - axis])
        result = progress_imap(
            partial(do_mad, workers_queue=workers_queue, axis=axis, skipna=skipna, level=level), tasks, workers_queue,
            n_cpu=n_cpu, disable=disable_pr_bar, show_vmem=show_vmem, total=total, desc='MAD'
        )
        return pd.concat(result)

    return p_mad


def do_idxmax(df, workers_queue, axis, skipna):
    def foo():
        return df.idxmax(axis=axis, skipna=skipna)

    return progress_udf_wrapper(foo, workers_queue, 1)()


def parallelize_idxmax(n_cpu=None, disable_pr_bar=False, show_vmem=False, split_factor=1):
    @doc(DOC, func='idxmax')
    def p_idxmax(data, axis=0, skipna=True):
        workers_queue = Manager().Queue()
        split_size = get_split_size(n_cpu, split_factor)
        tasks = get_split_data(data, axis, split_size)
        total = min(split_size, data.shape[1 - axis])
        result = progress_imap(
            partial(do_idxmax, workers_queue=workers_queue, axis=axis, skipna=skipna), tasks, workers_queue,
            n_cpu=n_cpu, disable=disable_pr_bar, show_vmem=show_vmem, total=total, desc='IDXMAX'
        )
        return pd.concat(result)

    return p_idxmax


def do_idxmin(df, workers_queue, axis, skipna):
    def foo():
        return df.idxmin(axis=axis, skipna=skipna)

    return progress_udf_wrapper(foo, workers_queue, 1)()


def parallelize_idxmin(n_cpu=None, disable_pr_bar=False, show_vmem=False, split_factor=1):
    @doc(DOC, func='idxmin')
    def p_idxmin(data, axis=0, skipna=True):
        workers_queue = Manager().Queue()
        split_size = get_split_size(n_cpu, split_factor)
        tasks = get_split_data(data, axis, split_size)
        total = min(split_size, data.shape[1 - axis])
        result = progress_imap(
            partial(do_idxmin, workers_queue=workers_queue, axis=axis, skipna=skipna), tasks, workers_queue,
            n_cpu=n_cpu, disable=disable_pr_bar, show_vmem=show_vmem, total=total, desc='IDXMIN'
        )
        return pd.concat(result)

    return p_idxmin


def do_rank(df, workers_queue, axis, method, numeric_only, na_option, ascending, pct):
    def foo():
        return df.rank(axis=axis, method=method, numeric_only=numeric_only, na_option=na_option, ascending=ascending,
                       pct=pct)

    return progress_udf_wrapper(foo, workers_queue, 1)()


def parallelize_rank(n_cpu=None, disable_pr_bar=False, split_factor=1,
                     show_vmem=False):
    @doc(DOC, func='rank')
    def p_rank(data, axis=0, method: str = "average", numeric_only=lib.no_default, na_option="keep", ascending=True,
               pct=False):
        workers_queue = Manager().Queue()
        split_size = get_split_size(n_cpu, split_factor)
        tasks = get_split_data(data, axis, split_size)
        total = min(split_size, data.shape[1 - axis])
        result = progress_imap(
            partial(do_rank, workers_queue=workers_queue, axis=axis, method=method, numeric_only=numeric_only,
                    na_option=na_option, ascending=ascending, pct=pct), tasks, workers_queue,
            n_cpu=n_cpu, disable=disable_pr_bar, show_vmem=show_vmem, total=total, desc='RANK'
        )
        return pd.concat(result, axis=1 - axis)

    return p_rank


def do_quantile(df, workers_queue, axis, q, numeric_only, interpolation):
    def foo():
        return df.quantile(axis=axis, q=q, numeric_only=numeric_only, interpolation=interpolation)

    return progress_udf_wrapper(foo, workers_queue, 1)()


def parallelize_quantile(n_cpu=None, disable_pr_bar=False, split_factor=1,
                         show_vmem=False):
    @doc(DOC, func='quantile')
    def p_quantile(data, q=0.5, axis=0, numeric_only: bool = True, interpolation: str = "linear"):
        workers_queue = Manager().Queue()
        split_size = get_split_size(n_cpu, split_factor)
        tasks = get_split_data(data, axis, split_size)
        total = min(split_size, data.shape[1 - axis])
        result = progress_imap(
            partial(do_quantile, workers_queue=workers_queue, axis=axis, numeric_only=numeric_only, q=q,
                    interpolation=interpolation), tasks, workers_queue,
            n_cpu=n_cpu, disable=disable_pr_bar, show_vmem=show_vmem, total=total, desc='QUANTILE'
        )
        if not lib.is_list_like(q):
            return pd.concat(result)
        return pd.concat(result, axis=1)

    return p_quantile


def do_mode(df, workers_queue, axis, numeric_only, dropna):
    def foo():
        return df.mode(axis=axis, numeric_only=numeric_only, dropna=dropna)

    return progress_udf_wrapper(foo, workers_queue, 1)()


def parallelize_mode(n_cpu=None, disable_pr_bar=False, split_factor=1,
                     show_vmem=False):
    @doc(DOC, func='mode')
    def p_mode(data, executor='processes', axis=0, numeric_only: bool = False, dropna=True):
        workers_queue = Manager().Queue()
        split_size = get_split_size(n_cpu, split_factor)
        tasks = get_split_data(data, axis, split_size)
        total = min(split_size, data.shape[1 - axis])
        result = progress_imap(
            partial(do_mode, workers_queue=workers_queue, axis=axis, numeric_only=numeric_only, dropna=dropna
                    ), tasks, workers_queue,
            n_cpu=n_cpu, disable=disable_pr_bar, show_vmem=show_vmem, total=total, executor=executor, desc='MODE'
        )
        return pd.concat(result, axis=1 - axis)

    return p_mode


def do_merge(df, workers_queue, right, **kwargs):
    def foo():
        return df.merge(right, **kwargs)

    return progress_udf_wrapper(foo, workers_queue, 1)()


def parallelize_merge(n_cpu=None, disable_pr_bar=False, split_factor=1,
                      show_vmem=False):
    @doc(DOC, func='mode')
    def p_merge(data,
                right: pd.DataFrame | pd.Series,
                how: str = "inner",
                on: IndexLabel | None = None,
                left_on: IndexLabel | None = None,
                right_on: IndexLabel | None = None,
                left_index: bool = False,
                right_index: bool = False,
                sort: bool = False,
                suffixes: Suffixes = ("_x", "_y"),
                copy: bool = True,
                indicator: bool = False,
                validate: str | None = None, ):
        workers_queue = Manager().Queue()
        split_size = get_split_size(n_cpu, split_factor)
        tasks = get_split_data(data, 1, split_size)
        total = min(split_size, data.shape[0])
        result = progress_imap(
            partial(do_merge, workers_queue=workers_queue, how=how, right=right, on=on, left_on=left_on, right_on=right_on,
                    left_index=left_index, right_index=right_index, sort=sort, suffixes=suffixes, copy=copy, indicator=indicator,
                    validate=validate), tasks, workers_queue, n_cpu=n_cpu, disable=disable_pr_bar, show_vmem=show_vmem,
            total=total, executor='threads', desc='MERGE'
        )
        return pd.concat(result)

    return p_merge


def do_isin(df, workers_queue, values):
    def foo():
        return df.isin(values)

    return progress_udf_wrapper(foo, workers_queue, 1)()


def parallelize_isin(n_cpu=None, disable_pr_bar=False, split_factor=1,
                     show_vmem=False):
    @doc(DOC, func='isin')
    def p_isin(data, values):
        workers_queue = Manager().Queue()
        split_size = get_split_size(n_cpu, split_factor)
        tasks = get_split_data(data, 1, split_size)
        total = min(split_size, data.shape[0])
        result = progress_imap(
            partial(do_isin, workers_queue=workers_queue, values=values), tasks, workers_queue,
            n_cpu=n_cpu, disable=disable_pr_bar, show_vmem=show_vmem, total=total, executor='threads', desc='ISIN'
        )
        return pd.concat(result)

    return p_isin


class ParallelizeStatFunc:
    def __init__(self, n_cpu=None, disable_pr_bar=False, show_vmem=False, split_factor=1):
        self.n_cpu = n_cpu
        self.disable_pr_bar = disable_pr_bar
        self.show_vmem = show_vmem
        self.split_factor = split_factor

    @staticmethod
    def get_nanops_arg(name):
        if name == 'min':
            return nanops.nanmin
        if name == 'max':
            return nanops.nanmax
        if name == 'mean':
            return nanops.nanmean
        if name == 'median':
            return nanops.nanmedian
        if name == 'skew':
            return nanops.nanskew
        if name == 'kurt':
            return nanops.nankurt

    def _stat_func(self, df, workers_queue, name, axis, skipna, level, numeric_only, kwargs):
        def closure():
            return df._stat_function(name, self.get_nanops_arg(name), axis, skipna, level, numeric_only, **kwargs)

        return progress_udf_wrapper(closure, workers_queue, 1)()

    def _parallel_stat_func(self, data, name, kwargs, axis=0, skipna=True, level=None, numeric_only=None):
        workers_queue = Manager().Queue()
        split_size = get_split_size(self.n_cpu, self.split_factor)
        tasks = get_split_data(data, axis, split_size)
        total = min(split_size, data.shape[1 - axis])
        result = progress_imap(
            partial(self._stat_func, workers_queue=workers_queue, name=name, axis=axis, skipna=skipna,
                    level=level, numeric_only=numeric_only, kwargs=kwargs), tasks, workers_queue,
            total=total, n_cpu=self.n_cpu, disable=self.disable_pr_bar, show_vmem=self.show_vmem, desc=name.upper())
        return pd.concat(result)

    def do_parallel(self, name):
        if name == 'min':
            @doc(DOC, func=name)
            def p_min(data, axis=0, skipna=True, level=None, numeric_only=None, **kwargs):
                return self._parallel_stat_func(data, name=name, axis=axis, skipna=skipna,
                                                level=level, numeric_only=numeric_only, kwargs=kwargs)

            return p_min
        if name == 'max':
            @doc(DOC, func=name)
            def p_max(data, axis=0, skipna=True, level=None, numeric_only=None, **kwargs):
                return self._parallel_stat_func(data, name=name, axis=axis, skipna=skipna,
                                                level=level, numeric_only=numeric_only, kwargs=kwargs)

            return p_max
        if name == 'mean':
            @doc(DOC, func=name)
            def p_mean(data, axis=0, skipna=True, level=None, numeric_only=None, **kwargs):
                return self._parallel_stat_func(data, name=name, axis=axis, skipna=skipna,
                                                level=level, numeric_only=numeric_only, kwargs=kwargs)

            return p_mean
        if name == 'median':
            @doc(DOC, func=name)
            def p_median(data, axis=0, skipna=True, level=None, numeric_only=None, **kwargs):
                return self._parallel_stat_func(data, name=name, axis=axis, skipna=skipna,
                                                level=level, numeric_only=numeric_only, kwargs=kwargs)

            return p_median
        if name == 'kurt':
            @doc(DOC, func=name)
            def p_kurt(data, axis=0, skipna=True, level=None, numeric_only=None, **kwargs):
                return self._parallel_stat_func(data, name=name, axis=axis, skipna=skipna,
                                                level=level, numeric_only=numeric_only, kwargs=kwargs)

            return p_kurt

        if name == 'skew':
            @doc(DOC, func=name)
            def p_skew(data, axis=0, skipna=True, level=None, numeric_only=None, **kwargs):
                return self._parallel_stat_func(data, name=name, axis=axis, skipna=skipna,
                                                level=level, numeric_only=numeric_only, kwargs=kwargs)

            return p_skew


class ParallelizeStatFuncDdof:
    def __init__(self, n_cpu=None, disable_pr_bar=False, show_vmem=False, split_factor=1):
        self.n_cpu = n_cpu
        self.disable_pr_bar = disable_pr_bar
        self.show_vmem = show_vmem
        self.split_factor = split_factor

    @staticmethod
    def get_nanops_arg(name):
        if name == 'sem':
            return nanops.nansem
        if name == 'var':
            return nanops.nanvar
        if name == 'std':
            return nanops.nanstd

    def _stat_func_ddof(self, df, workers_queue, name, axis, skipna, level, ddof, numeric_only, kwargs):
        def closure():
            return df._stat_function_ddof(name, self.get_nanops_arg(name), axis, skipna, level, ddof, numeric_only,
                                          **kwargs)

        return progress_udf_wrapper(closure, workers_queue, 1)()

    def _parallel_stat_func_ddof(self, data, name, kwargs, axis=0, skipna=True, level=None, ddof=1,
                                 numeric_only=None):
        workers_queue = Manager().Queue()
        split_size = get_split_size(self.n_cpu, self.split_factor)
        tasks = get_split_data(data, axis, split_size)
        total = min(split_size, data.shape[1 - axis])
        result = progress_imap(
            partial(self._stat_func_ddof, workers_queue=workers_queue, name=name, axis=axis, skipna=skipna,
                    level=level, ddof=ddof, numeric_only=numeric_only, kwargs=kwargs), tasks, workers_queue,
            total=total, n_cpu=self.n_cpu, disable=self.disable_pr_bar, show_vmem=self.show_vmem, desc=name.upper())
        return pd.concat(result)

    def do_parallel(self, name):
        if name == 'var':
            @doc(DOC, func=name)
            def p_var(data, axis=0, skipna=True, level=None, ddof=1, numeric_only=None, **kwargs):
                return self._parallel_stat_func_ddof(data, name=name, axis=axis,
                                                     skipna=skipna, level=level, ddof=ddof, numeric_only=numeric_only,
                                                     kwargs=kwargs)

            return p_var
        if name == 'std':
            @doc(DOC, func=name)
            def p_std(data, axis=0, skipna=True, level=None, ddof=1, numeric_only=None, **kwargs):
                return self._parallel_stat_func_ddof(data, name=name, axis=axis,
                                                     skipna=skipna, level=level, ddof=ddof, numeric_only=numeric_only,
                                                     kwargs=kwargs)

            return p_std
        if name == 'sem':
            @doc(DOC, func=name)
            def p_sem(data, axis=0, skipna=True, level=None, ddof=1, numeric_only=None, **kwargs):
                return self._parallel_stat_func_ddof(data, name=name, axis=axis,
                                                     skipna=skipna, level=level, ddof=ddof, numeric_only=numeric_only,
                                                     kwargs=kwargs)

            return p_sem


class ParallelizeMinCountStatFunc:
    def __init__(self, n_cpu=None, disable_pr_bar=False, show_vmem=False, split_factor=1):
        self.n_cpu = n_cpu
        self.disable_pr_bar = disable_pr_bar
        self.show_vmem = show_vmem
        self.split_factor = split_factor

    @staticmethod
    def get_nanops_arg(name):
        if name == 'sum':
            return nanops.nansum
        if name == 'prod':
            return nanops.nanprod

    def _min_count_stat_func(self, df, workers_queue, name, axis, skipna, level, numeric_only, min_count, kwargs):
        def closure():
            return df._min_count_stat_function(name, self.get_nanops_arg(name), axis, skipna, level, numeric_only,
                                               min_count, **kwargs
                                               )

        return progress_udf_wrapper(closure, workers_queue, 1)()

    def _parallel_min_count_stat_func(self, data, name, kwargs, axis=0, skipna=True, level=None,
                                      numeric_only=None, min_count=0):
        workers_queue = Manager().Queue()
        split_size = get_split_size(self.n_cpu, self.split_factor)
        tasks = get_split_data(data, axis, split_size)
        total = min(split_size, data.shape[1 - axis])
        result = progress_imap(
            partial(self._min_count_stat_func, workers_queue=workers_queue, name=name, axis=axis, skipna=skipna,
                    level=level, min_count=min_count, numeric_only=numeric_only, kwargs=kwargs), tasks, workers_queue,
            total=total, n_cpu=self.n_cpu, disable=self.disable_pr_bar, show_vmem=self.show_vmem, desc=name.upper())
        return pd.concat(result)

    def do_parallel(self, name):
        if name == 'sum':
            @doc(DOC, func=name)
            def p_sum(data, axis=0, skipna=True, level=None, numeric_only=None,
                      min_count=0, **kwargs):
                return self._parallel_min_count_stat_func(data, name=name, axis=axis,
                                                          skipna=skipna, level=level, min_count=min_count,
                                                          numeric_only=numeric_only, kwargs=kwargs)

            return p_sum
        if name == 'prod':
            @doc(DOC, func=name)
            def p_prod(data, axis=0, skipna=True, level=None, numeric_only=None,
                       min_count=0, **kwargs):
                return self._parallel_min_count_stat_func(data, name=name, axis=axis,
                                                          skipna=skipna, level=level, min_count=min_count,
                                                          numeric_only=numeric_only, kwargs=kwargs)

            return p_prod


class ParallelizeAccumFunc:
    def __init__(self, n_cpu=None, disable_pr_bar=False, show_vmem=False, split_factor=1):
        self.n_cpu = n_cpu
        self.disable_pr_bar = disable_pr_bar
        self.show_vmem = show_vmem
        self.split_factor = split_factor

    @staticmethod
    def get_func(name):
        if name == 'cummin':
            return np.minimum.accumulate
        if name == 'cummax':
            return np.maximum.accumulate
        if name == 'cumsum':
            return np.cumsum
        if name == 'cumprod':
            return np.cumprod

    def _concat_by_columns(self, columns):
        first = columns[0]
        for d in columns[1:]:
            first.update(d)
        return pd.DataFrame(first)

    def _accum_func(self, df, workers_queue, name, axis, skipna, args, kwargs):
        def closure():
            if not axis:
                return df._accum_func(name, self.get_func(name), axis, skipna, *args, **kwargs).to_dict(orient='series')
            return df._accum_func(name, self.get_func(name), axis, skipna, *args, **kwargs)

        return progress_udf_wrapper(closure, workers_queue, 1)()

    def _parallel_accum_func(self, data, name, args, kwargs, axis=0, skipna=True):
        workers_queue = Manager().Queue()
        split_size = get_split_size(self.n_cpu, self.split_factor)
        tasks = get_split_data(data, axis, split_size)
        total = min(split_size, data.shape[1 - axis])
        result = progress_imap(
            partial(self._accum_func, workers_queue=workers_queue, name=name, axis=axis, skipna=skipna,
                    args=args, kwargs=kwargs), tasks, workers_queue,
            total=total, n_cpu=self.n_cpu, disable=self.disable_pr_bar, show_vmem=self.show_vmem, desc=name.upper())
        if not axis:
            return self._concat_by_columns(result)
        return pd.concat(result, axis=1 - axis)

    def do_parallel(self, name):
        if name == 'cumsum':
            @doc(DOC, func=name)
            def p_cumsum(data, axis=0, skipna=True, *args, **kwargs):
                return self._parallel_accum_func(data, name=name, axis=axis,
                                                 skipna=skipna, args=args, kwargs=kwargs)

            return p_cumsum

        if name == 'cumprod':
            @doc(DOC, func=name)
            def p_cumprod(data, axis=0, skipna=True, *args, **kwargs):
                return self._parallel_accum_func(data, name=name, axis=axis,
                                                 skipna=skipna, args=args, kwargs=kwargs)

            return p_cumprod

        if name == 'cummin':
            @doc(DOC, func=name)
            def p_cummin(data, axis=0, skipna=True, *args, **kwargs):
                return self._parallel_accum_func(data, name=name, axis=axis,
                                                 skipna=skipna, args=args, kwargs=kwargs)

            return p_cummin

        if name == 'cummax':
            @doc(DOC, func=name)
            def p_cummax(data, axis=0, skipna=True, *args, **kwargs):
                return self._parallel_accum_func(data, name=name, axis=axis,
                                                 skipna=skipna, args=args, kwargs=kwargs)

            return p_cummax



from setuptools import setup, find_packages

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='parallel-pandas',
    python_requires='>=3.7',
    version='0.6.1',
    packages=find_packages(),
    author='Dubovik Pavel',
    author_email='geometryk@gmail.com',
    description='Parallel processing on pandas with progress bars',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=[
        'parallel pandas',
        'progress bar',
        'parallel apply',
        'parallel groupby',
        'multiprocessing bar',
    ],
    url='https://github.com/dubovikmaster/parallel-pandas',
    license='MIT',
    install_requires=[
        'pandas >= 1.4.0',
        'dill',
        'psutil',
        'tqdm',
        'scipy',
    ],
    extras_require={
        'numba': [
            'numba',
        ]
    },
    platforms='any',
)



import numpy as np

try:
    from numba import njit, prange
    @njit
    def _pearson_corr(x, y):
        return np.corrcoef(x, y)[0, 1]


    @njit
    def calculate_rank(arr):
        n = len(arr)
        result = np.zeros(n, dtype=np.int64)
        indexes = np.argsort(arr)
        for i in range(n):
            result[indexes[i]] = i + 1

        return result


    @njit(fastmath=True)
    def _calculate_rank_dot_product(a, b):
        n = len(a)
        b_rank = calculate_rank(b)
        result = 0.0
        indexes = np.argsort(a)
        for i in range(n):
            result += (i + 1 - b_rank[indexes[i]]) * (i + 1 - b_rank[indexes[i]])

        return result


    @njit
    def _spearman_corr_with_rank(x, y):
        n = len(x)
        sum_d_squared = _calculate_rank_dot_product(x, y)
        correlation = 1 - (6 * sum_d_squared) / (n * (n ** 2 - 1))

        return correlation


    @njit(parallel=True)
    def _parallel_spearman_corr(mat, min_periods):
        n = mat.shape[0]
        c = mat.shape[1]
        rows, cols = np.triu_indices(c, k=1)
        corr_mat = np.zeros((c, c), dtype=np.float64)
        ranks = np.zeros((n, c), dtype=np.int64)
        isnan_flags = np.zeros(c, dtype=np.bool_)
        isnan = np.zeros((n, c), dtype=np.bool_)
        for i in prange(c):
            isnan[:, i] = np.isnan(mat[:, i])
            if isnan[:, i].any():
                isnan_flags[i] = True
            #
            else:
                ranks[:, i] = calculate_rank(mat[:, i])
        for i in prange(len(rows)):
            row, col = rows[i], cols[i]
            if isnan_flags[row] or isnan_flags[col]:
                valid = ~isnan[:, row] & ~isnan[:, col]
                if valid.sum() < min_periods:
                    corr_mat[row, col] = np.nan
                else:
                    corr_mat[row, col] = _spearman_corr_with_rank(mat[valid, row], mat[valid, col])
            else:
                d = ranks[:, row] - ranks[:, col]
                sum_d_squared = np.sum(d * d)
                corr_mat[row, col] = 1 - (6 * sum_d_squared) / (n * (n ** 2 - 1))
        corr_mat = corr_mat + corr_mat.transpose()
        np.fill_diagonal(corr_mat, 1)
        return corr_mat


    @njit(parallel=True)
    def _parallel_pearson_corr(mat, min_periods):
        n = mat.shape[0]
        c = mat.shape[1]
        rows, cols = np.triu_indices(c, k=1)
        corr_mat = np.zeros((mat.shape[1], mat.shape[1]), dtype=np.float64)
        isnan = np.isnan(mat)
        isnan_flags = isnan.sum(axis=0)
        sums = np.sum(mat, axis=0)
        sum_squareds = np.sum((mat * mat), axis=0)
        for i in prange(len(rows)):
            row, col = rows[i], cols[i]
            if isnan_flags[row] or isnan_flags[col]:
                valid = ~isnan[:, row] & ~isnan[:, col]
                if valid.sum() < min_periods:
                    corr_mat[row, col] = np.nan
                else:
                    corr_mat[row, col] = _pearson_corr(mat[valid, row], mat[valid, col])
            else:
                sum_xy = np.sum(mat[:, row] * mat[:, col])
                numerator = n * sum_xy - sums[row] * sums[col]
                denominator = np.sqrt(
                    (n * sum_squareds[row] - sums[row] * sums[row]) * (n * sum_squareds[col] - sums[col] * sums[col]))
                corr_mat[row, col] = numerator / denominator if denominator != 0 else 0
        corr_mat = corr_mat + corr_mat.transpose()
        np.fill_diagonal(corr_mat, 1)
        return corr_mat


    @njit(fastmath=True)
    def _kendall_tau(x, y):
        concordant = 0
        discordant = 0
        n = len(x)
        for i in range(n - 1):
            for j in range(i + 1, n):
                # Проверка согласованности или расхождения порядка
                if (x[i] - x[j]) * (y[i] - y[j]) > 0:
                    concordant += 1
                elif (x[i] - x[j]) * (y[i] - y[j]) < 0:
                    discordant += 1
        # Вычисление корреляции Кендалла
        tau = (concordant - discordant) / np.sqrt((concordant + discordant) * (n * (n - 1) / 2))

        return tau


    @njit(parallel=True)
    def _parallel_kendall(mat, min_periods):
        n = mat.shape[0]
        c = mat.shape[1]
        rows, cols = np.triu_indices(c, k=1)
        corr_mat = np.zeros((c, c), dtype=np.float64)
        isnan = np.isnan(mat)
        isnan_flags = isnan.sum(axis=0)
        for i in prange(len(rows)):
            row, col = rows[i], cols[i]
            a = mat[:, row]
            b = mat[:, col]
            if isnan_flags[row] or isnan_flags[col]:
                valid = ~isnan[:, row] & ~isnan[:, col]
                if valid.sum() < min_periods:
                    corr_mat[row, col] = np.nan
                else:
                    corr_mat[row, col] = _kendall_tau(a[valid], b[valid])
            else:
                corr_mat[row, col] = _kendall_tau(a, b)
        corr_mat = corr_mat + corr_mat.transpose()
        np.fill_diagonal(corr_mat, 1)
        return corr_mat


    @njit
    def _do_parallel_corr(mat, method, min_periods):
        if method == 'pearson':
            return _parallel_pearson_corr(mat, min_periods)
        elif method == 'spearman':
            return _parallel_spearman_corr(mat, min_periods)
        elif method == 'kendall':
            return _parallel_kendall(mat, min_periods)
        else:
            raise ValueError(f'Unknown method {method}')

except ImportError as e:
    def _do_parallel_corr(*args, **kwargs):
        raise ImportError('Numba not installed. Please install numba to use parallel pandas with numba engine.')



import pandas as pd

from .core import series_parallelize_apply
from .core import series_parallelize_map
from .core import parallelize_apply
from .core import parallelize_chunk_apply
from .core import parallelize_replace
from .core import ParallelizeStatFunc
from .core import ParallelizeStatFuncDdof
from .core import parallelize_groupby_apply
from .core import parallelize_applymap
from .core import parallelize_describe
from .core import parallelize_nunique
from .core import parallelize_mad
from .core import parallelize_idxmax
from .core import parallelize_idxmin
from .core import parallelize_rank
from .core import ParallelizeMinCountStatFunc
from .core import ParallelizeAccumFunc
from .core import parallelize_quantile
from .core import parallelize_mode
from .core import parallelize_isin
from .core import parallelize_aggregate
from .core import ParallelRolling
from .core import ParallelExpanding
from .core import ParallelEWM
from .core import ParallelEWMGroupby
from .core import ParallelExpandingGroupby
from .core import ParallelRollingGroupby
from .core import ParallelWindow
from .core import parallelize_corr

ROLL_AND_EXP_OPS = ['mean', 'max', 'min', 'sum', 'std', 'var', 'median', 'skew', 'kurt', 'sem', 'quantile', 'rank',
                    'apply', 'aggregate', 'agg']
EWM_OPS = WIN_OPS = ['mean', 'sum', 'std', 'var']


class ParallelPandas:
    @staticmethod
    def initialize(n_cpu=None, disable_pr_bar=False, show_vmem=False, split_factor=1):
        # add parallel methods to Series
        pd.Series.p_apply = series_parallelize_apply(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar, show_vmem=show_vmem,
                                                     split_factor=split_factor)
        pd.Series.p_map = series_parallelize_map(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar, show_vmem=show_vmem,
                                                 split_factor=split_factor)
        pd.Series.p_isin = parallelize_isin(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar, show_vmem=show_vmem,
                                            split_factor=split_factor)

        # add parallel methods to DataFrame
        pd.DataFrame.p_corr = parallelize_corr(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar, show_vmem=show_vmem,
                                               split_factor=split_factor)
        pd.DataFrame.p_apply = parallelize_apply(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar, show_vmem=show_vmem,
                                                 split_factor=split_factor)

        pd.DataFrame.p_agg = parallelize_aggregate(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar, show_vmem=show_vmem,
                                                   split_factor=split_factor)

        pd.DataFrame.p_replace = parallelize_replace(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                     show_vmem=show_vmem, split_factor=split_factor)

        pd.DataFrame.p_min = ParallelizeStatFunc(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                 show_vmem=show_vmem, split_factor=split_factor).do_parallel('min')

        pd.DataFrame.p_max = ParallelizeStatFunc(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                 show_vmem=show_vmem, split_factor=split_factor).do_parallel('max')

        pd.DataFrame.p_mean = ParallelizeStatFunc(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                  show_vmem=show_vmem, split_factor=split_factor).do_parallel('mean')
        pd.DataFrame.p_median = ParallelizeStatFunc(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                    show_vmem=show_vmem,
                                                    split_factor=split_factor).do_parallel('median')
        pd.DataFrame.p_skew = ParallelizeStatFunc(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                  show_vmem=show_vmem, split_factor=split_factor).do_parallel('skew')

        pd.DataFrame.p_kurt = ParallelizeStatFunc(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                  show_vmem=show_vmem, split_factor=split_factor).do_parallel('kurt')

        pd.DataFrame.p_std = ParallelizeStatFuncDdof(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                     show_vmem=show_vmem, split_factor=split_factor).do_parallel('std')

        pd.DataFrame.p_var = ParallelizeStatFuncDdof(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                     show_vmem=show_vmem, split_factor=split_factor).do_parallel('var')

        pd.DataFrame.p_sem = ParallelizeStatFuncDdof(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                     show_vmem=show_vmem, split_factor=split_factor).do_parallel('sem')

        pd.DataFrame.p_sum = ParallelizeMinCountStatFunc(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                         show_vmem=show_vmem,
                                                         split_factor=split_factor).do_parallel('sum')

        pd.DataFrame.p_prod = ParallelizeMinCountStatFunc(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                          show_vmem=show_vmem,
                                                          split_factor=split_factor).do_parallel('prod')

        pd.DataFrame.p_cumprod = ParallelizeAccumFunc(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                      show_vmem=show_vmem,
                                                      split_factor=split_factor).do_parallel('cumprod')

        pd.DataFrame.p_cummin = ParallelizeAccumFunc(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                     show_vmem=show_vmem,
                                                     split_factor=split_factor).do_parallel('cummin')

        pd.DataFrame.p_cummax = ParallelizeAccumFunc(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                     show_vmem=show_vmem,
                                                     split_factor=split_factor).do_parallel('cummax')

        pd.DataFrame.p_cumsum = ParallelizeAccumFunc(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                     show_vmem=show_vmem,
                                                     split_factor=split_factor).do_parallel('cumsum')

        pd.DataFrame.p_applymap = parallelize_applymap(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                       show_vmem=show_vmem)
        pd.DataFrame.p_describe = parallelize_describe(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                       show_vmem=show_vmem,
                                                       split_factor=split_factor)

        pd.DataFrame.p_nunique = parallelize_nunique(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                     show_vmem=show_vmem,
                                                     split_factor=split_factor)

        pd.DataFrame.p_mad = parallelize_mad(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar, show_vmem=show_vmem,
                                             split_factor=split_factor)

        pd.DataFrame.p_idxmax = parallelize_idxmax(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar, show_vmem=show_vmem,
                                                   split_factor=split_factor)

        pd.DataFrame.p_idxmin = parallelize_idxmin(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar, show_vmem=show_vmem,
                                                   split_factor=split_factor)

        pd.DataFrame.p_rank = parallelize_rank(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar, show_vmem=show_vmem,
                                               split_factor=split_factor)

        pd.DataFrame.p_quantile = parallelize_quantile(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar, show_vmem=show_vmem,
                                                       split_factor=split_factor)

        pd.DataFrame.p_mode = parallelize_mode(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar, show_vmem=show_vmem,
                                               split_factor=split_factor)

        pd.DataFrame.chunk_apply = parallelize_chunk_apply(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                           show_vmem=show_vmem,
                                                           split_factor=split_factor)

        pd.DataFrame.p_isin = parallelize_isin(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                               show_vmem=show_vmem,
                                               split_factor=split_factor)

        # Rolling parallel methods
        for name in ROLL_AND_EXP_OPS:
            setattr(pd.core.window.Rolling, 'p_' + name, ParallelRolling(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                                         show_vmem=show_vmem,
                                                                         split_factor=split_factor).do_parallel(name))

        # Window parallel methods
        for name in WIN_OPS:
            setattr(pd.core.window.Window, 'p_' + name, ParallelWindow(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                                       show_vmem=show_vmem,
                                                                       split_factor=split_factor).do_parallel(
                name))

        # Expanding parallel methods
        for name in ROLL_AND_EXP_OPS:
            setattr(pd.core.window.Expanding, 'p_' + name, ParallelExpanding(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                                                             show_vmem=show_vmem,
                                                                             split_factor=split_factor).do_parallel(
                name))

        # ExponentialMovingWindow parallel methods
        for name in EWM_OPS:
            setattr(pd.core.window.ExponentialMovingWindow, 'p_' + name,
                    ParallelEWM(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                show_vmem=show_vmem,
                                split_factor=split_factor).do_parallel(
                        name))

        # RollingGroupby parallel methods
        for name in ROLL_AND_EXP_OPS:
            setattr(pd.core.window.RollingGroupby, 'p_' + name,
                    ParallelRollingGroupby(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                           show_vmem=show_vmem,
                                           split_factor=split_factor).do_parallel(name))

        # ExpandingGroupby parallel methods
        for name in ROLL_AND_EXP_OPS:
            setattr(pd.core.window.ExpandingGroupby, 'p_' + name,
                    ParallelExpandingGroupby(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                             show_vmem=show_vmem,
                                             split_factor=split_factor).do_parallel(
                        name))

        # ExponentialMovingWindow parallel methods
        for name in EWM_OPS:
            setattr(pd.core.window.ExponentialMovingWindowGroupby, 'p_' + name,
                    ParallelEWMGroupby(n_cpu=n_cpu, disable_pr_bar=disable_pr_bar,
                                       show_vmem=show_vmem,
                                       split_factor=split_factor).do_parallel(
                        name))

        # add parallel methods to DataFrameGroupBy and SeriesGroupBy
        pd.core.groupby.DataFrameGroupBy.p_apply = parallelize_groupby_apply(n_cpu=n_cpu,
                                                                             disable_pr_bar=disable_pr_bar)
        pd.core.groupby.SeriesGroupBy.p_apply = parallelize_groupby_apply(n_cpu=n_cpu,
                                                                          disable_pr_bar=disable_pr_bar)
